# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
from typing import Optional, Any, Dict, List, Tuple, Literal

import datetime
import logging
from pathlib import Path
from copy import deepcopy
from collections import OrderedDict

import networkx

from canals.errors import PipelineConnectError, PipelineMaxLoops, PipelineRuntimeError, PipelineValidationError
from canals.component.input_output import ComponentInput
from canals.draw import draw, RenderingEngines
from canals.pipeline._utils import (
    InputSocket,
    OutputSocket,
    find_input_sockets,
    find_output_sockets,
    find_unambiguous_connection,
    parse_connection_name,
    validate_pipeline_input,
)


logger = logging.getLogger(__name__)


class Pipeline:
    """
    Components orchestration engine.

    Builds a graph of components and orchestrates their execution according to the execution graph.
    """

    def __init__(
        self,
        metadata: Optional[Dict[str, Any]] = None,
        max_loops_allowed: int = 100,
    ):
        """
        Creates the Pipeline.

        Args:
            metadata: arbitrary dictionary to store metadata about this pipeline. Make sure all the values contained in
                this dictionary can be serialized and deserialized if you wish to save this pipeline to file with
                `save_pipelines()/load_pipelines()`.
            max_loops_allowed: how many times the pipeline can run the same node before throwing an exception.
        """
        self.metadata = metadata or {}
        self.max_loops_allowed = max_loops_allowed
        self.graph = networkx.MultiDiGraph()
        self.debug: Dict[int, Dict[str, Any]] = {}

    def __eq__(self, other) -> bool:
        """
        Equal pipelines share every metadata, node and edge, but they're not required to use
        the same node instances: this allows pipeline saved and then loaded back to be equal to themselves.
        """
        if (
            not isinstance(other, type(self))
            or not getattr(self, "metadata") == getattr(other, "metadata")
            or not getattr(self, "max_loops_allowed") == getattr(other, "max_loops_allowed")
            or not hasattr(self, "graph")
            or not hasattr(other, "graph")
        ):
            return False

        return (
            self.graph.adj == other.graph.adj
            and self._comparable_nodes_list(self.graph) == self._comparable_nodes_list(other.graph)
            and self.graph.graph == other.graph.graph
        )

    def _comparable_nodes_list(self, graph: networkx.MultiDiGraph) -> List[Dict[str, Any]]:
        """
        Replaces instances of nodes with their class name and defaults list in order to make sure they're comparable.
        """
        nodes = []
        for node in graph.nodes:
            comparable_node = graph.nodes[node]
            if hasattr(comparable_node, "defaults"):
                comparable_node["defaults"] = comparable_node["instance"].defaults
            comparable_node["instance"] = comparable_node["instance"].__class__
            nodes.append(comparable_node)
        nodes.sort()
        return nodes

    def add_component(self, name: str, instance: Any) -> None:
        """
        Create a component for the given component. Components are not connected to anything by default:
        use `Pipeline.connect()` to connect components together.

        Component names must be unique, but component instances can be reused if needed.

        Args:
            name: the name of the component.
            instance: the component instance.

        Returns:
            None

        Raises:
            ValueError: if a component with the same name already exists
            PipelineValidationError: if the given instance is not a Canals component
        """
        # Component names are unique
        if name in self.graph.nodes:
            raise ValueError(f"A component named '{name}' already exists in this pipeline: choose another name.")

        # Components can't be named `_debug`
        if name == "_debug":
            raise ValueError("'_debug' is a reserved name for debug output. Choose another name.")

        # Component instances must be components
        if not hasattr(instance, "__canals_component__"):
            raise PipelineValidationError(
                f"'{type(instance)}' doesn't seem to be a component. Is this class decorated with @component?"
            )

        # Find inputs and outputs
        input_sockets = find_input_sockets(instance)
        output_sockets = find_output_sockets(instance)

        # Add component to the graph, disconnected
        logger.debug("Adding component '%s' (%s)", name, instance)
        self.graph.add_node(
            name,
            instance=instance,
            input_sockets=input_sockets,
            variadic_input=any(e.variadic for e in input_sockets.values()),
            output_sockets=output_sockets,
            visits=0,
        )

    def connect(self, connect_from: str, connect_to: str) -> None:
        """
        Connects two components together. All components to connect must exist in the pipeline.
        If connecting to an component that has several output connections, specify the inputs and output names as
        'component_name.connections_name'.

        Args:
            connect_from: the component that delivers the value. This can be either just a component name or can be
                in the format `component_name.connection_name` if the component has multiple outputs.
            connect_to: the component that receives the value. This can be either just a component name or can be
                in the format `component_name.connection_name` if the component has multiple inputs.

        Returns:
            None

        Raises:
            PipelineConnectError: if the two components cannot be connected (for example if one of the components is
                not present in the pipeline, or the connections don't match by type, and so on).
        """
        # Edges may be named explicitly by passing 'node_name.edge_name' to connect().
        from_node, from_socket_name = parse_connection_name(connect_from)
        to_node, to_socket_name = parse_connection_name(connect_to)

        # Get the nodes data.
        try:
            from_sockets = self.graph.nodes[from_node]["output_sockets"]
        except KeyError as exc:
            raise ValueError(f"Component named {from_node} not found in the pipeline.") from exc

        try:
            to_sockets = self.graph.nodes[to_node]["input_sockets"]
        except KeyError as exc:
            raise ValueError(f"Component named {to_node} not found in the pipeline.") from exc

        # If the name of either socket is given, get the socket
        if from_socket_name:
            from_socket = from_sockets.get(from_socket_name, None)
            if not from_socket:
                raise PipelineConnectError(
                    f"'{from_node}.{from_socket_name} does not exist. "
                    f"Output connections of {from_node} are: "
                    + ", ".join([f"{name} (type {socket.type.__name__})" for name, socket in from_sockets.items()])
                )
        if to_socket_name:
            to_socket = to_sockets.get(to_socket_name, None)
            if not to_socket:
                raise PipelineConnectError(
                    f"'{to_node}.{to_socket_name} does not exist. "
                    f"Input connections of {to_node} are: "
                    + ", ".join([f"{name} (type {socket.type.__name__})" for name, socket in to_sockets.items()])
                )

        # If either one of the two sockets is not specified, look for an unambiguous connection
        # Note that if there is more than one possible connection but two sockets match by name, they're paired.
        if not to_socket_name or not from_socket_name:
            from_sockets = [from_socket] if from_socket_name else from_sockets.values()
            to_sockets = [to_socket] if to_socket_name else to_sockets.values()
            from_socket, to_socket = find_unambiguous_connection(
                from_node=from_node, from_sockets=from_sockets, to_node=to_node, to_sockets=to_sockets
            )

        # Connect the components on these sockets
        self._direct_connect(from_node=from_node, from_socket=from_socket, to_node=to_node, to_socket=to_socket)

    def _direct_connect(self, from_node: str, from_socket: OutputSocket, to_node: str, to_socket: InputSocket) -> None:
        """
        Directly connect socket to socket.
        """
        # Check that the types match. We need type equality: subclass relationships are not accepted, just like
        # Optionals, Unions, and similar "aggregate" types. See https://github.com/python/typing/issues/570
        if not from_socket.type == to_socket.type:
            raise PipelineConnectError(
                f"Cannot connect '{from_node}.{from_socket.name}' with '{to_node}.{to_socket.name}': "
                f"their declared input and output types do not match.\n"
                f" - {from_node}.{from_socket.name}: {from_socket.type.__name__}\n"
                f" - {to_node}.{to_socket.name}: {to_socket.type.__name__}\n"
            )

        # Make sure the receiving socket is not taken - sending sockets can be connected as many times as needed,
        # so they don't need this check
        if to_socket.taken_by:
            raise PipelineConnectError(
                f"Cannot connect '{from_node}.{from_socket.name}' with '{to_node}.{to_socket.name}': "
                f"{to_node}.{to_socket.name} is already connected to {to_socket.taken_by}.\n"
            )

        # Create the connection
        logger.debug("Connecting '%s.%s' to '%s.%s'", from_node, from_socket.name, to_node, to_socket.name)
        edge_key = f"{from_socket.name}/{to_socket.name}"
        self.graph.add_edge(from_node, to_node, key=edge_key, from_socket=from_socket, to_socket=to_socket)

        # Mark the receiving socket as taken (unless is variadic - variadic sockets are never "taken")
        if not to_socket.variadic:
            to_socket.taken_by = from_node

    def get_component(self, name: str) -> object:
        """
        Returns an instance of a component.

        Args:
            name: the name of the component

        Returns:
            The instance of that component.

        Raises:
            ValueError: if a component with that name is not present in the pipeline.
        """
        try:
            return self.graph.nodes[name]["instance"]
        except KeyError as exc:
            raise ValueError(f"Component named {name} not found in the pipeline.") from exc

    def draw(self, path: Path, engine: RenderingEngines = "mermaid-img") -> None:
        """
        Draws the pipeline. Requires either `graphviz` as a system dependency, or an internet connection for Mermaid.
        Run `pip install canals[graphviz]` or `pip install canals[mermaid]` to install missing dependencies.

        Args:
            path: where to save the diagram.
            engine: which format to save the graph as. Accepts 'graphviz', 'mermaid-text', 'mermaid-img'.
                Default is 'mermaid-img'.

        Returns:
            None

        Raises:
            ImportError: if `engine='graphviz'` and `pygraphviz` is not installed.
            HTTPConnectionError: (and similar) if the internet connection is down or other connection issues.
        """
        draw(graph=deepcopy(self.graph), path=path, engine=engine)

    def warm_up(self):
        """
        Make sure all nodes are warm.

        It's the node's responsibility to make sure this method can be called at every `Pipeline.run()`
        without re-initializing everything.
        """
        for node in self.graph.nodes:
            if hasattr(self.graph.nodes[node]["instance"], "warm_up"):
                logger.info("Warming up component %s...", node)
                self.graph.nodes[node]["instance"].warm_up()

    def run(self, data: Dict[str, ComponentInput], debug: bool = False) -> Dict[str, Any]:
        """
        Runs the pipeline.

        Args:
            data: the inputs to give to the input components of the Pipeline.
            parameters: a dictionary with all the parameters of all the components, namespaced by component.
            debug: whether to collect and return debug information.

        Returns:
            A dictionary with the outputs of the output components of the Pipeline.

        Raises:
            PipelineRuntimeError: if the any of the components fail or return unexpected output.
        """
        # **** The Pipeline.run() algorithm ****
        #
        # Nodes are run as soon as an input for them appears in the inputs buffer.
        # When there's more than a node at once in the buffer (which means some
        # branches are running in parallel or that there are loops) they are selected to
        # run in FIFO order by the `inputs_buffer` OrderedDict.
        #
        # Inputs are labeled with the name of the node they're aimed for:
        #
        #   ````
        #   inputs_buffer[target_node] = {"input_name": input_value, ...}
        #   ```
        #
        # Nodes should wait until all the necessary input data has arrived before running.
        # If they're popped from the input_buffer before they're ready, they're put back in.
        # If the pipeline has branches of different lengths, it's possible that a node has to
        # wait a bit and let other nodes pass before receiving all the input data it needs.
        #
        # Chetsheet for networkx data access:
        # - Name of the node       # self.graph.nodes  (List[str])
        # - Node instance          # self.graph.nodes[node]["instance"]
        # - Input nodes            # [e[0] for e in self.graph.in_edges(node)]
        # - Output nodes           # [e[1] for e in self.graph.out_edges(node)]
        # - Output edges           # [e[2]["label"] for e in self.graph.out_edges(node, data=True)]
        #
        # if debug:
        #     os.makedirs("debug", exist_ok=True)

        data = validate_pipeline_input(self.graph, input_values=data)
        self._clear_visits_count()
        self.warm_up()

        logger.info("Pipeline execution started.")
        inputs_buffer = OrderedDict(
            {
                node: {key: value for key, value in input_data.__dict__.items() if value is not None}
                for node, input_data in data.items()
            }
        )
        pipeline_output: Dict[str, Dict[str, Any]] = {}

        if debug:
            logger.info("Debug mode ON.")
            self.debug = {}

        # *** PIPELINE EXECUTION LOOP ***
        # We select the nodes to run by popping them in FIFO order from the inputs buffer.
        step = 0
        while inputs_buffer:
            step += 1
            if debug:
                self.debug[step] = {
                    "time": datetime.datetime.now(),
                    "inputs_buffer": list(inputs_buffer.items()),
                    "pipeline_output": pipeline_output,
                }
            logger.debug("> Queue at step %s: %s", step, {k: list(v.keys()) for k, v in inputs_buffer.items()})

            component, inputs = inputs_buffer.popitem(last=False)  # FIFO

            # if debug:
            #     draw(deepcopy(self.graph), engine="graphviz", path=f"debug/step_{current_step}.jpg", running=component, queued=inputs_buffer.keys())

            # **** IS IT MY TURN YET? ****
            # Check if the node should be run or not
            ready_to_run = self._ready_to_run(name=component, inputs=inputs, inputs_buffer=inputs_buffer)

            # This component is missing data: let's put it back in the queue and wait.
            if ready_to_run == "wait":
                inputs_buffer[component] = inputs
                continue

            # This component did not receive the input it needs: it must be on a skipped branch. Let's not run it.
            if ready_to_run == "skip":
                self.graph.nodes[component]["visits"] += 1
                inputs_buffer = self._skip_downstream_nodes(component=component, inputs_buffer=inputs_buffer)
                continue

            # **** RUN THE NODE ****
            # It is our turn! The node is ready to run and all necessary inputs are present
            output = self._run_component(name=component, inputs=inputs)

            # **** PROCESS THE OUTPUT ****
            # The node run successfully. Let's store or distribute the output it produced, if it's valid.
            if not self.graph.out_edges(component):
                # Note: if a node outputs many times (like in loops), the output will be overwritten
                pipeline_output[component] = output
            else:
                inputs_buffer = self._route_output(
                    node_results=output, node_name=component, inputs_buffer=inputs_buffer
                )

        if debug:
            self.debug[step + 1] = {
                "time": datetime.datetime.now(),
                "inputs_buffer": list(inputs_buffer.items()),
                "pipeline_output": pipeline_output,
            }
            pipeline_output["_debug"] = self.debug  # type: ignore

        logger.info("Pipeline executed successfully.")
        return pipeline_output

    def _clear_visits_count(self):
        """
        Make sure all nodes's visits count is zero.
        """
        for node in self.graph.nodes:
            self.graph.nodes[node]["visits"] = 0

    def _check_max_loops(self, component_name: str):
        """
        Verify whether this component run too many times.
        """
        if self.graph.nodes[component_name]["visits"] > self.max_loops_allowed:
            raise PipelineMaxLoops(
                f"Maximum loops count ({self.max_loops_allowed}) exceeded for component '{component_name}'."
            )

    def _ready_to_run(
        self, name: str, inputs: Dict[str, Any], inputs_buffer: OrderedDict
    ) -> Literal["run", "wait", "skip"]:
        """
        Verify whether a component is ready to run.

        Returns 'run', 'wait' or 'skip' depending on how the node should be treated and the log message explaining
        the decision.
        """
        # Make sure it didn't run too many times already
        self._check_max_loops(name)

        # List all the component/socket pairs the current component should be waiting for.
        expected_inputs = self._connections_to_wait_for(name=name)

        # Check if the expected inputs were all received
        if self._check_received_vs_expected_inputs(name=name, inputs=inputs, expected_inputs=expected_inputs):
            return "run"

        # This node is missing some inputs. Did all the upstream nodes run?
        nodes_to_wait_for, _ = zip(*expected_inputs) if expected_inputs else ([], [])

        # Some node upstream didn't run yet, so we should wait for them.
        if not self._all_nodes_to_wait_for_run(nodes_to_wait_for=nodes_to_wait_for):

            if not inputs_buffer:
                # What if there are no components to wait for?
                raise PipelineRuntimeError(
                    f"'{name}' is stuck waiting for input, but there are no other components to run. "
                    "This is likely a Canals bug. Open an issue at https://github.com/deepset-ai/canals."
                )

            logger.debug(
                "Putting '%s' back in the queue, some inputs are missing (inputs to wait for: %s, inputs_received: %s)",
                name,
                [f"{node}.{socket}" for node, socket in expected_inputs],
                list(inputs.keys()),
            )
            return "wait"

        # All upstream nodes run, so it **must** be our turn.
        # However we're missing data, so this branch probably is being skipped.
        if inputs and self.graph.nodes[name]["variadic_input"]:
            logger.debug(
                "Running '%s', even though some upstream component did not produced output. "
                "(upstream components: %s, expected inputs: %s, n. of inputs received %s)",
                name,
                list(nodes_to_wait_for),
                [f"{node}.{socket}" for node, socket in expected_inputs],
                len(list(inputs.values())[0]) if inputs else 0,
            )
            return "run"

        logger.debug(
            "Skipping '%s', upstream component didn't produce output "
            "(upstream components: %s, expected inputs: %s, inputs received: %s)",
            name,
            list(nodes_to_wait_for),
            [f"{node}.{socket}" for node, socket in expected_inputs],
            list(inputs.keys()),
        )
        return "skip"

    def _check_received_vs_expected_inputs(
        self, name: str, inputs: Dict[str, Any], expected_inputs: Tuple[str, str]
    ) -> bool:
        """
        Check if all the inputs the component is expecting have been received.

        Returns True if all the necessary inputs are received, False otherwise, and a message with the decision.
        """
        # Variadic nodes expect a single list regardless of how many incoming connections they have,
        # but the length of the list should match the length of incoming connections.
        if self.graph.nodes[name]["variadic_input"]:

            # Variadic nodes need at least two values
            if not inputs or len(inputs) < 2:
                return False

            if len(list(inputs.values())[0]) == len(expected_inputs):
                logger.debug(
                    "Component '%s' is ready to run: all connected inputs were received "
                    "(expecting %s, received %s values).",
                    name,
                    len(expected_inputs),
                    len(list(inputs.values())[0]),
                )
                return True
        else:
            # No input sockets are connected: this is an input node and should be always ready to run.
            if not expected_inputs:
                logger.debug("Component '%s' is ready to run: it's a starting node.", name)
                return True

            # Otherwise, just make sure there is one input key for each expected input key
            _, expected_input_names = zip(*expected_inputs)
            if set(expected_input_names).issubset(set(inputs.keys())):
                logger.debug("Component '%s' is ready to run: all expected inputs were received.", name)
                return True

        return False

    def _connections_to_wait_for(self, name: str):
        """
        Return all the component/socket pairs this component is waiting for.
        """
        # We should be wait on all edges except for the downstream ones, to support loops.
        # This downstream check is enabled only for nodes taking more than one input
        # (the "entrance" of the loop).
        data_to_wait_for = [
            (from_node, data["to_socket"].name)
            for from_node, _, data in self.graph.in_edges(name, data=True)
            # ... if there's a path in the graph leading back from the current node to the
            # input node, # and only in case this node accepts multiple inputs.
            if not (networkx.has_path(self.graph, name, from_node) and self.graph.nodes[name]["variadic_input"])
        ]
        return data_to_wait_for

    def _all_nodes_to_wait_for_run(self, nodes_to_wait_for: List[str]) -> bool:
        """
        Check if all the nodes this component is waiting for has run or not.

        FIXME: checking for `visits>0` might not be enough for all loops.
        """
        return all(self.graph.nodes[node_to_wait_for]["visits"] > 0 for node_to_wait_for in nodes_to_wait_for)

    def _skip_downstream_nodes(self, component: str, inputs_buffer: OrderedDict) -> OrderedDict:
        """
        When a component is skipped, put all downstream nodes in the inputs buffer too: the might be skipped too,
        unless they are merge/variadic nodes. They will be evaluated later by the pipeline execution loop.
        """
        downstream_nodes = [e[1] for e in self.graph.out_edges(component)]
        for downstream_node in downstream_nodes:
            if not downstream_node in inputs_buffer:
                inputs_buffer[downstream_node] = {}
        return inputs_buffer

    def _run_component(self, name: str, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """
        Once we're confident this component is ready to run, run it and collect the output.
        """
        self.graph.nodes[name]["visits"] += 1
        instance = self.graph.nodes[name]["instance"]
        try:
            logger.info("* Running %s (visits: %s)", name, self.graph.nodes[name]["visits"])
            logger.debug("   '%s' inputs: %s", name, inputs)

            input_class = instance.Input if hasattr(instance, "Input") else instance.input_type

            # If the node is variadic, unpack the input
            if self.graph.nodes[name]["variadic_input"]:
                inputs = list(inputs.values())[0]
                input_dataclass = input_class(*inputs)

            # Otherwise pass the inputs as kwargs after adding the component's own defaults to them
            else:
                inputs = {**instance.defaults, **inputs}
                input_dataclass = input_class(**inputs)

            output_dataclass = instance.run(input_dataclass)

            # Unwrap the output
            logger.debug("   '%s' outputs: %s\n", name, output_dataclass.__dict__)

        except Exception as e:
            raise PipelineRuntimeError(
                f"{name} raised '{e.__class__.__name__}: {e}' \nInputs: {inputs}\n\n"
                "See the stacktrace above for more information."
            ) from e

        return output_dataclass

    def _route_output(
        self,
        node_name: str,
        node_results: Dict[str, Any],
        inputs_buffer: OrderedDict,
    ) -> OrderedDict:
        """
        Distrubute the outputs of the component into the input buffer of downstream components.

        Returns the updated inputs buffer.
        """
        # This is not a terminal node: find out where the output goes, to which nodes and along which edge
        is_decision_node_for_loop = (
            any(networkx.has_path(self.graph, edge[1], node_name) for edge in self.graph.out_edges(node_name))
            and len(self.graph.out_edges(node_name)) > 1
        )
        for edge_data in self.graph.out_edges(node_name, data=True):
            to_socket = edge_data[2]["to_socket"]
            from_socket = edge_data[2]["from_socket"]
            target_node = edge_data[1]

            # If this is a decision node and a loop is involved, we add to the input buffer only the nodes
            # that received their expected output and we leave the others out of the queue.
            if is_decision_node_for_loop and getattr(node_results, from_socket.name) is None:
                if networkx.has_path(self.graph, target_node, node_name):
                    # In case we're choosing to leave a loop, do not put the loop's node in the buffer.
                    logger.debug(
                        "Not adding '%s' to the inputs buffer: we're leaving the loop.",
                        target_node,
                    )
                else:
                    # In case we're choosing to stay in a loop, do not put the external node in the buffer.
                    logger.debug(
                        "Not adding '%s' to the inputs buffer: we're staying in the loop.",
                        target_node,
                    )
            else:
                # In all other cases, populate the inputs buffer for all downstream nodes, setting None to any
                # edge that did not receive input.
                if not target_node in inputs_buffer:
                    inputs_buffer[target_node] = {}  # Create the buffer for the downstream node if it's not there yet

                value_to_route = getattr(node_results, from_socket.name, None)
                if value_to_route:
                    if to_socket.variadic:
                        if not to_socket.name in inputs_buffer[target_node].keys():
                            inputs_buffer[target_node][to_socket.name] = []
                        inputs_buffer[target_node][to_socket.name].append(value_to_route)
                    else:
                        inputs_buffer[target_node][to_socket.name] = value_to_route

        return inputs_buffer



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
from typing import TypeVar, Any, List

from dataclasses import make_dataclass, dataclass

import pytest

from canals.testing import BaseTestComponent
from canals.component import component, ComponentInput, ComponentOutput


@component
class Repeat:
    """
    Repeats the input value on all outputs.
    """

    @dataclass
    class Input(ComponentInput):
        value: int

    def __init__(self, outputs: List[str] = ["output_1", "output_2", "output_3"]):
        self.outputs = outputs
        self._output_type = make_dataclass(
            "Output", fields=[(val, int, None) for val in outputs], bases=(ComponentOutput,)
        )

    @property
    def output_type(self):
        return self._output_type

    def run(self, data: Input):
        output_dataclass = self.output_type()
        for output in self.outputs:
            setattr(output_dataclass, output, data.value)
        return output_dataclass


class TestRepeat(BaseTestComponent):
    @pytest.fixture
    def components(self):
        return [Repeat(), Repeat(outputs=["one", "two"])]

    def test_repeat_default(self):
        component = Repeat()
        results = component.run(Repeat.Input(value=10))
        assert results == component.output_type(output_1=10, output_2=10, output_3=10)
        assert component.init_parameters == {}

    def test_repeat_init(self):
        component = Repeat(outputs=["one", "two"])
        results = component.run(Repeat.Input(value=10))
        assert results == component.output_type(one=10, two=10)
        assert component.init_parameters == {"outputs": ["one", "two"]}



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
import logging
import inspect

from canals.errors import ComponentError
from canals.component.decorators import save_init_params, init_defaults


logger = logging.getLogger(__name__)


def component(class_):
    """
    Marks a class as a component. Any class decorated with `@component` can be used by a Pipeline.

    All components must follow the contract below. This docstring is the source of truth for components contract.

    ### `@component` decorator

    All component classes must be decorated with the `@component` decorator. This allows Canals to discover them.

    ### `Input`

    ```python
    @dataclass
    class Input(ComponentInput / VariadicComponentInput):
        <expected input fields, typed, with no defaults>
    ```
    Semi-mandatory method (either this or `self.input_type(self)`).

    This inner class defines how the input of this component looks like. For example, if the node is expecting
    a list of Documents, the fields of the class should be `documents: List[Document]`

    Defaults are allowed, however `Optional`, `Union` and similar "generic" types are not. This is necessary to allow
    proper validation of the connections, which rely on the type of these fields.

    If your node expects variadic input, use `VariadicComponentInput`. In all other scenarios, use `ComponentInput`
    as your base class.

    Some components may need more dynamic input. For these scenarios, refer to `self.input_type()`.

    Every component should define **either** `Input` or `self.input_type()`.


    ### `input_type()`

    ```python
    @property
    def input_type(self) -> ComponentInput / VariadicComponentInput:
    ```
    Semi-mandatory method (either this or `class Input`).

    This method defines how the input of this component looks like. For example, if the node is expecting
    a list of Documents, this method should return a dataclass, subclass of either `ComponentInput` or
    `VariadicComponentInput`, with such fields. For example, it could build the dataclass as
    `make_dataclass("Input", fields=[(f"documents", List[Document], None)], bases=(ComponentInput, ))` and return it.

    Defaults are allowed, however `Optional`, `Union` and similar "generic" types are not. This is necessary to allow
    proper validation of the connections, which rely on the type of these fields.

    Normally the `Input` dataclass is preferred, as it provides autocompletion for the users and is much easier to use.

    Every component should define **either** `Input` or `self.input_type()`.


    ### `Output`

    ```python
    @dataclass
    class Output(ComponentOutput):
        <expected output fields, typed>
    ```
    Semi-mandatory method (either this or `self.output_type()`).

    This inner class defines how the output of this component looks like. For example, if the node is producing
    a list of Documents, the fields of the class should be `documents: List[Document]`

    Defaults are allowed, however `Optional`, `Union` and similar "generic" types are not. This is necessary to allow
    proper validation of the connections, which rely on the type of these fields.

    Some components may need more dynamic output: for example, your component accepts a list of file extensions at
    init time and wants to have one output field for each of those. For these scenarios, refer to `self.output_type()`.

    Every component should define **either** `Output` or `self.output_type()`.


    ### `output_type()`

    ```python
    @property
    def output_type(self) -> ComponentOutput:
    ```
    Semi-mandatory method (either this or `class Output`).

    This method defines how the output of this component looks like. For example, if the node is producing
    a list of Documents, this method should return a dataclass with such fields, for example:
    `return make_dataclass("Output", fields=[(f"documents", List[Document], None)], bases=(ComponentOutput, ))`

    Defaults are allowed, however `Optional`, `Union` and similar "generic" types are not. This is necessary to allow
    proper validation of the connections, which rely on the type of these fields.

    If the output is static, normally the `Output` dataclass is preferred, as it provides autocompletion for the users.

    Every component should define **either** `Output` or `self.output_type`.


    ### `__init__()`

    ```python
    def __init__(self, [... components init parameters ...]):
    ```
    Optional method.

    Components may have an `__init__` method where they define:

    - `self.defaults = {parameter_name: parameter_default_value, ...}`:
        All values defined here will be sent to the `run()` method when the Pipeline calls it.
        If any of these parameters is also receiving input from other components, those have precedence.
        This collection of values is supposed to replace the need for default values in `run()` and make them
        dynamically configurable. Keep in mind that only these defaults will count at runtime: defaults given to
        the `Input` dataclass (see above) will be ignored.

    - `self.init_parameters = {same parameters that the __init__ method received}`:
        In this dictionary you can store any state the components wish to be persisted when they are saved.
        These values will be given to the `__init__` method of a new instance when the pipeline is loaded.
        Note that by default the `@component` decorator saves the arguments automatically.
        However, if a component sets their own `init_parameters` manually in `__init__()`, that will be used instead.
        Note: all of the values contained here **must be JSON serializable**. Serialize them manually if needed.

    Components should take only "basic" Python types as parameters of their `__init__` function, or iterables and
    dictionaries containing only such values. Anything else (objects, functions, etc) will raise an exception at init
    time. If there's the need for such values, consider serializing them to a string.

    _(TODO explain how to use classes and functions in init. In the meantime see `test/components/test_accumulate.py`)_

    The `__init__` must be extrememly lightweight, because it's a frequent operation during the construction and
    validation of the pipeline. If a component has some heavy state to initialize (models, backends, etc...) refer to
    the `warm_up()` method.


    ### `warm_up()`

    ```python
    def warm_up(self):
    ```
    Optional method.

    This method is called by Pipeline before the graph execution. Make sure to avoid double-initializations,
    because Pipeline will not keep track of which components it called `warm_up()` on.


    ### `run()`

    ```python
    def run(self, data: <Input if defined, otherwise untyped>) -> <Output if defined, otherwise untyped>:
    ```
    Mandatory method.

    This is the method where the main functionality of the component should be carried out. It's called by
    `Pipeline.run()`.

    When the component should run, Pipeline will call this method with:

    - all the input values coming from other components connected to it,
    - if any is missing, the corresponding value defined in `self.defaults`, if it exists.

    `run()` must return a single instance of the dataclass declared through either `Output` or `self.output_type()`.

    Args:
        class_: the class that Canals should use as a component.
        serializable: whether to check, at init time, if the component can be saved with
        `save_pipelines()`.

    Returns:
        A class that can be recognized as a component.

    Raises:
        ComponentError: if the class provided has no `run()` method or otherwise doesn't respect the component contract.
    """
    logger.debug("Registering %s as a component", class_)

    # '__canals_component__' is used to distinguish components from regular classes.
    # Its value is set to the desired component name: normally it is the class name, but it can technically be customized.
    class_.__canals_component__ = class_.__name__

    # Check for Input
    if not hasattr(class_, "Input") and not hasattr(class_, "input_type"):
        raise ComponentError(
            "Components must either have an Input dataclass or a 'input_type' property that returns such dataclass"
        )
    if hasattr(class_, "Input") and not hasattr(class_.Input, "_component_input"):
        raise ComponentError(f"{class_.__name__}.Input must inherit from ComponentInput")

    # Check for Output
    if not hasattr(class_, "Output") and not hasattr(class_, "output_type"):
        raise ComponentError(
            "Components must either have an Output dataclass or a 'output_type' property that returns such dataclass"
        )
    if hasattr(class_, "Output") and not hasattr(class_.Output, "_component_output"):
        raise ComponentError(f"{class_.__name__}.Output must inherit from ComponentOutput")

    # Check for run()
    if not hasattr(class_, "run"):
        raise ComponentError(f"{class_.__name__} must have a 'run()' method. See the docs for more information.")
    run_signature = inspect.signature(class_.run)

    # run() must take a single input param
    if len(run_signature.parameters) != 2:
        raise ComponentError("run() must accept only a single parameter called 'data'.")

    # The input param must be called data
    if not "data" in run_signature.parameters:
        raise ComponentError("run() must accept a parameter called 'data'.")

    # Either give a self.input_type function or type 'data' with the Input dataclass
    if not hasattr(class_, "input_type") and run_signature.parameters["data"].annotation != class_.Input:
        raise ComponentError(f"'data' must be typed and the type must be {class_.__name__}.Input.")

    # Check for the return types
    if not hasattr(class_, "output_type") and run_signature.return_annotation == inspect.Parameter.empty:
        raise ComponentError(f"{class_.__name__}.run() must declare the type of its return value.")

    # Automatically registers all the init parameters in an instance attribute called `init_parameters`.
    # See `save_init_params()`.
    class_.__init__ = save_init_params(class_.__init__)

    # Makes sure the self.defaults dictionary is always present
    class_.__init__ = init_defaults(class_.__init__)

    return class_



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
from pathlib import Path
from pprint import pprint

from canals.pipeline import Pipeline
from test.sample_components import AddFixedValue, Subtract

import logging

logging.basicConfig(level=logging.DEBUG)


def test_pipeline(tmp_path):

    add_two = AddFixedValue(add=2)
    diff = Subtract()

    pipeline = Pipeline()
    pipeline.add_component("first_addition", add_two)
    pipeline.add_component("second_addition", add_two)
    pipeline.add_component("third_addition", add_two)
    pipeline.add_component("diff", diff)
    pipeline.add_component("fourth_addition", AddFixedValue(add=1))

    pipeline.connect("first_addition", "second_addition")
    pipeline.connect("second_addition", "diff.first_value")
    pipeline.connect("third_addition", "diff.second_value")
    pipeline.connect("diff", "fourth_addition.value")

    pipeline.draw(tmp_path / "fixed_merging_pipeline.png")

    results = pipeline.run(
        {
            "first_addition": AddFixedValue.Input(value=1),
            "third_addition": AddFixedValue.Input(value=1),
        }
    )
    pprint(results)

    assert results == {"fourth_addition": AddFixedValue.Output(value=3)}


if __name__ == "__main__":
    test_pipeline(Path(__file__).parent)



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0

# pylint: disable=protected-access

import logging
import inspect
from functools import wraps


logger = logging.getLogger(__name__)


def save_init_params(init_func):
    """
    Decorator that saves the init parameters of a component in `self.init_parameters`
    """

    @wraps(init_func)
    def wrapper_saveinit_parameters(self, *args, **kwargs):

        # Call the actual __init__ function with the arguments
        init_func(self, *args, **kwargs)

        # Convert all args into kwargs
        sig = inspect.signature(init_func)
        arg_names = list(sig.parameters.keys())
        if any(arg_names) and arg_names[0] in ["self", "cls"]:
            arg_names.pop(0)
        args_as_kwargs = {arg_name: arg for arg, arg_name in zip(args, arg_names)}

        # Collect and store all the init parameters, preserving whatever the components might have already added there
        init_parameters = {**args_as_kwargs, **kwargs}
        if hasattr(self, "init_parameters"):
            init_parameters = {**init_parameters, **self.init_parameters}
        self.init_parameters = init_parameters

    return wrapper_saveinit_parameters


def init_defaults(init_func):
    """
    Decorator that makes sure the `self.defaults` dictionary exists
    """

    @wraps(init_func)
    def wrapper_create_defaults_dict(self, *args, **kwargs):

        # Call the actual __init__ function with the arguments
        init_func(self, *args, **kwargs)

        # Makes sure the component has a defaults dictionary the pipeline can use
        if not hasattr(self, "defaults"):
            self.defaults = {}

    return wrapper_create_defaults_dict



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
from typing import Optional

from dataclasses import dataclass

import pytest

from canals.testing import BaseTestComponent
from canals.component import component, ComponentInput, ComponentOutput


@component
class Threshold:
    """
    Redirects the value, unchanged, along a different connection whether the value is above
    or below the given threshold.

    Single input, double output decision component.

    :param threshold: the number to compare the input value against. This is also a parameter.
    """

    @dataclass
    class Input(ComponentInput):
        value: int
        threshold: int = 10

    @dataclass
    class Output(ComponentOutput):
        above: int
        below: int

    def __init__(self, threshold: Optional[int] = None):
        """
        :param threshold: the number to compare the input value against.
        """
        if threshold:
            self.defaults = {"threshold": threshold}

    def run(self, data: Input) -> Output:
        if data.value < data.threshold:
            return Threshold.Output(above=None, below=data.value)  # type: ignore
        return Threshold.Output(above=data.value, below=None)  # type: ignore


class TestThreshold(BaseTestComponent):
    @pytest.fixture
    def components(self):
        return [Threshold()]

    def test_threshold(self):
        component = Threshold()

        results = component.run(Threshold.Input(value=5, threshold=10))
        assert results == Threshold.Output(above=None, below=5)

        results = component.run(Threshold.Input(value=15, threshold=10))
        assert results == Threshold.Output(above=15, below=None)



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
from dataclasses import dataclass

import pytest

from canals.testing import BaseTestComponent
from canals.component import component, ComponentInput, ComponentOutput


@component
class Subtract:
    """
    Compute the difference between two values.
    """

    @dataclass
    class Input(ComponentInput):
        first_value: int
        second_value: int

    @dataclass
    class Output(ComponentOutput):
        difference: int

    def run(self, data: Input) -> Output:
        """
        :param first_value: name of the connection carrying the value to subtract from.
        :param second_value: name of the connection carrying the value to subtract.
        """
        return Subtract.Output(difference=data.first_value - data.second_value)


class TestSubtract(BaseTestComponent):
    @pytest.fixture
    def components(self):
        return [Subtract()]

    def test_subtract(self):
        component = Subtract()
        results = component.run(Subtract.Input(first_value=10, second_value=7))
        assert results == Subtract.Output(difference=3)
        assert component.init_parameters == {}



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
from typing import List

from dataclasses import dataclass

import pytest

from canals.testing import BaseTestComponent
from canals.component import component, VariadicComponentInput, ComponentOutput


@component
class Sum:
    """
    Sums the values of all the input connections together.
    """

    @dataclass
    class Input(VariadicComponentInput):
        values: List[int]

    @dataclass
    class Output(ComponentOutput):
        total: int

    def run(self, data: Input) -> Output:
        return Sum.Output(total=sum(data.values))


class TestSum(BaseTestComponent):
    @pytest.fixture
    def components(self):
        return [Sum()]

    def test_sum_no_values(self):
        component = Sum()
        results = component.run(Sum.Input())
        assert results == Sum.Output(total=0)
        assert component.init_parameters == {}

    def test_sum_one_value(self):
        component = Sum()
        results = component.run(Sum.Input(10))
        assert results == Sum.Output(total=10)
        assert component.init_parameters == {}

    def test_sum_few_values(self):
        component = Sum()
        results = component.run(Sum.Input(10, 11, 12))
        assert results == Sum.Output(total=33)
        assert component.init_parameters == {}



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
from canals.pipeline.pipeline import Pipeline
from canals.errors import (
    PipelineError,
    PipelineRuntimeError,
    PipelineValidationError,
)
from canals.pipeline.save_load import (
    save_pipelines,
    load_pipelines,
    marshal_pipelines,
    unmarshal_pipelines,
    _find_decorated_classes,
)



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
from typing import List

from dataclasses import dataclass

import pytest

from canals.component import component, VariadicComponentInput, ComponentOutput
from canals.testing import BaseTestComponent


@component
class MergeLoop:
    """
    Takes two input components and returns the first one that is not None.
    In case both received a value, priority is given to 'first'.
    """

    def __init__(self, expected_type: type):
        self.expected_type = expected_type
        self.init_parameters = {"expected_type": expected_type.__name__}

    @property
    def input_type(self):
        @dataclass
        class Input(VariadicComponentInput):
            values: List[self.expected_type]  # type: ignore

        return Input

    @property
    def output_type(self):
        @dataclass
        class Output(ComponentOutput):
            value: self.expected_type  # type: ignore

        return Output

    def run(self, data):
        """
        Takes some inputs and returns the first one that is not None.

        In case it receives more than one value, priority is given to the first.
        """
        for v in data.values:
            if v is not None:
                return self.output_type(value=v)
        return self.output_type(value=None)


class TestMergeLoop(BaseTestComponent):
    @pytest.fixture
    def components(self):

        # Serialization is broken :'(
        return [
            # MergeLoop(expected_type=int),
            # MergeLoop(expected_type=str),
            # MergeLoop(expected_type=List[str])
        ]

    def test_merge_first(self):
        component = MergeLoop(expected_type=int)
        results = component.run(component.input_type(5, None))
        assert results.__dict__ == component.output_type(value=5).__dict__

    def test_merge_second(self):
        component = MergeLoop(expected_type=int)
        results = component.run(component.input_type(None, 5))
        assert results.__dict__ == component.output_type(value=5).__dict__

    def test_merge_nones(self):
        component = MergeLoop(expected_type=int)
        results = component.run(component.input_type(None, None, None))
        assert results.__dict__ == component.output_type(value=None).__dict__

    def test_merge_one(self):
        component = MergeLoop(expected_type=int)
        results = component.run(component.input_type(1))
        assert results.__dict__ == component.output_type(value=1).__dict__

    def test_merge_one_none(self):
        component = MergeLoop(expected_type=int)
        results = component.run(component.input_type())
        assert results.__dict__ == component.output_type(value=None).__dict__



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
import logging
import inspect
from dataclasses import fields

logger = logging.getLogger(__name__)


class BaseIODataclass:
    def names(self):
        """
        Returns the name of all the fields of this dataclass.
        """
        return [field.name for field in fields(self)]

    def to_dict(self):
        """
        Returns a dictionary representation of this dataclass.
        """
        return self.__dict__


class Optionalize(type):
    """
    Makes all the fields of the dataclass optional by setting None as their default value.
    """

    def __call__(cls, *args, **kwargs):
        obj = cls.__new__(cls, *args, **kwargs)
        obj.__init__.__func__.__defaults__ = tuple(None for _ in inspect.signature(obj.__init__).parameters)
        obj.__init__(*args, **kwargs)
        return obj


class Variadic(type):
    """
    Adds the proper checks to a variadic input dataclass and packs the args into a list for the `__init__` call.
    """

    def __call__(cls, *args, **kwargs):
        if kwargs:
            raise ValueError(f"{cls.__name__} accepts only an unnamed list of positional parameters.")

        obj = cls.__new__(cls, *args)

        if len(inspect.signature(obj.__init__).parameters) != 1:
            raise ValueError(f"{cls.__name__} accepts only one variadic positional parameter.")

        obj.__init__(list(args))
        return obj


class ComponentInput(BaseIODataclass, metaclass=Optionalize):
    """
    Represents the input of a component.
    """

    # dataclasses are uncooperative (don't call `super()`), so we need this flag to check for inheritance
    _component_input = True


class VariadicComponentInput(BaseIODataclass, metaclass=Variadic):
    """
    Represents the input of a variadic component.
    """

    # VariadicComponentInput can't inherit from ComponentInput due to metaclasses clashes
    # dataclasses are uncooperative (don't call `super()`), so we need this flag to check for inheritance
    _component_input = True
    _variadic_component_input = True


class ComponentOutput(BaseIODataclass, metaclass=Optionalize):
    """
    Represents the output of a component.
    """

    # dataclasses are uncooperative (don't call `super()`), so we need this flag to check for inheritance
    _component_output = True



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
from pathlib import Path
from pprint import pprint
import logging

from canals.pipeline import Pipeline
from test.sample_components import (
    Accumulate,
    AddFixedValue,
    Greet,
    Parity,
    Threshold,
    Double,
    Sum,
    Repeat,
    Subtract,
    MergeLoop,
)

logging.basicConfig(level=logging.DEBUG)


def test_complex_pipeline(tmp_path):
    accumulate = Accumulate()

    pipeline = Pipeline(max_loops_allowed=4)
    pipeline.add_component("greet_first", Greet(message="Hello, the value is {value}."))
    pipeline.add_component("accumulate_1", accumulate)
    pipeline.add_component("add_two", AddFixedValue(add=2))
    pipeline.add_component("parity", Parity())
    pipeline.add_component("add_one", AddFixedValue(add=1))
    pipeline.add_component("accumulate_2", accumulate)

    pipeline.add_component("loop_merger", MergeLoop(expected_type=int))
    pipeline.add_component("below_10", Threshold(threshold=10))
    pipeline.add_component("double", Double())

    pipeline.add_component("greet_again", Greet(message="Hello again, now the value is {value}."))
    pipeline.add_component("sum", Sum())

    pipeline.add_component("greet_enumerator", Greet(message="Hello from enumerator, here the value became {value}."))
    pipeline.add_component("enumerate", Repeat(outputs=["first", "second"]))
    pipeline.add_component("add_three", AddFixedValue(add=3))

    pipeline.add_component("diff", Subtract())
    pipeline.add_component("greet_one_last_time", Greet(message="Bye bye! The value here is {value}!"))
    pipeline.add_component("replicate", Repeat(outputs=["first", "second"]))
    pipeline.add_component("add_five", AddFixedValue(add=5))
    pipeline.add_component("add_four", AddFixedValue(add=4))
    pipeline.add_component("accumulate_3", accumulate)

    pipeline.connect("greet_first", "accumulate_1")
    pipeline.connect("accumulate_1", "add_two")
    pipeline.connect("add_two", "parity")

    pipeline.connect("parity.even", "greet_again")
    pipeline.connect("greet_again", "sum")
    pipeline.connect("sum", "diff.first_value")
    pipeline.connect("diff", "greet_one_last_time")
    pipeline.connect("greet_one_last_time", "replicate")
    pipeline.connect("replicate.first", "add_five.value")
    pipeline.connect("replicate.second", "add_four.value")
    pipeline.connect("add_four", "accumulate_3")

    pipeline.connect("parity.odd", "add_one.value")
    pipeline.connect("add_one", "loop_merger")
    pipeline.connect("loop_merger", "below_10")

    pipeline.connect("below_10.below", "double")
    pipeline.connect("double", "loop_merger")

    pipeline.connect("below_10.above", "accumulate_2")
    pipeline.connect("accumulate_2", "diff.second_value")

    pipeline.connect("greet_enumerator", "enumerate")
    pipeline.connect("enumerate.second", "sum")

    pipeline.connect("enumerate.first", "add_three.value")
    pipeline.connect("add_three", "sum")

    pipeline.draw(tmp_path / "complex_pipeline.png")

    results = pipeline.run({"greet_first": Greet.Input(value=1), "greet_enumerator": Greet.Input(value=1)})
    pprint(results)
    print("accumulated: ", accumulate.state)

    assert results == {"accumulate_3": Accumulate.Output(value=9), "add_five": AddFixedValue.Output(value=-7)}
    assert accumulate.state == 9


if __name__ == "__main__":
    test_complex_pipeline(Path(__file__).parent)



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
from typing import Optional
import logging

from dataclasses import dataclass

import pytest

from canals.testing import BaseTestComponent
from canals.component import component, ComponentInput, ComponentOutput


logger = logging.getLogger(__name__)


@component
class Greet:
    """
    Logs a greeting message without affecting the value passing on the connection.
    """

    @dataclass
    class Input(ComponentInput):
        value: int
        message: str
        log_level: str

    @dataclass
    class Output(ComponentOutput):
        value: int

    def __init__(
        self,
        message: Optional[str] = "\nGreeting component says: Hi! The value is {value}\n",
        log_level: Optional[str] = "INFO",
    ):
        """
        :param message: the message to log. Can use `{value}` to embed the value.
        :param log_level: the level to log at.
        """
        if log_level and not getattr(logging, log_level):
            raise ValueError(f"This log level does not exist: {log_level}")
        self.defaults = {"message": message, "log_level": log_level}

    def run(self, data: Input) -> Output:
        """
        Logs a greeting message without affecting the value passing on the connection.
        """
        print(data.log_level)
        level = getattr(logging, data.log_level, None)
        if not level:
            raise ValueError(f"This log level does not exist: {data.log_level}")

        logger.log(level=level, msg=data.message.format(value=data.value))
        return Greet.Output(value=data.value)


class TestGreet(BaseTestComponent):
    @pytest.fixture
    def components(self):
        return [
            Greet(),
            Greet(message="Hello, that's {value}"),
            Greet(log_level="WARNING"),
            Greet(message="Hello, that's {value}", log_level="WARNING"),
        ]

    def test_greet_message(self, caplog):
        caplog.set_level(logging.WARNING)
        component = Greet()
        results = component.run(Greet.Input(value=10, message="Hello, that's {value}", log_level="WARNING"))
        assert results == Greet.Output(value=10)
        assert "Hello, that's 10" in caplog.text



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
from dataclasses import dataclass, make_dataclass

import pytest

from canals.testing import BaseTestComponent
from canals.component import component, ComponentInput, ComponentOutput


@component
class Remainder:
    """
    Redirects the value, unchanged, along the connection corresponding to the remainder
    of a division. For example, if `divisor=3`, the value `5` would be sent along
    the second output connection.
    """

    @dataclass
    class Input(ComponentInput):
        value: int
        add: int = 1

    def __init__(self, divisor: int = 2):
        if divisor == 0:
            raise ValueError("Can't divide by zero")
        self.divisor = divisor

        self._output_type = make_dataclass(
            "Output", fields=[(f"remainder_is_{val}", int, None) for val in range(divisor)], bases=(ComponentOutput,)
        )

    @property
    def output_type(self):
        return self._output_type

    def run(self, data: Input):
        """
        :param value: the value to check the remainder of.
        """
        remainder = data.value % self.divisor
        output = self.output_type()
        setattr(output, f"remainder_is_{remainder}", data.value)
        return output


class TestRemainder(BaseTestComponent):
    @pytest.fixture
    def components(self):
        return [Remainder(), Remainder(divisor=1)]

    def test_remainder_default(self):
        component = Remainder()
        results = component.run(Remainder.Input(value=3))
        assert results == component.output_type(remainder_is_1=3)

    def test_remainder_with_divisor(self):
        component = Remainder(divisor=4)
        results = component.run(Remainder.Input(value=3))
        assert results == component.output_type(remainder_is_3=3)

    def test_remainder_zero(self):
        with pytest.raises(ValueError):
            Remainder(divisor=0)



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
import pytest

from canals.pipeline import Pipeline, marshal_pipelines, unmarshal_pipelines
from test.sample_components import AddFixedValue, Double

import logging

logging.basicConfig(level=logging.DEBUG)


def test_marshal():
    add_1 = AddFixedValue(add=200)
    add_2 = AddFixedValue()

    pipeline_1 = Pipeline(metadata={"type": "test pipeline", "author": "me"})
    pipeline_1.add_component("first_addition", add_1)
    pipeline_1.add_component("double", Double())
    pipeline_1.add_component("second_addition", add_2)
    pipeline_1.add_component("third_addition", add_1)

    pipeline_1.connect("first_addition", "double")
    pipeline_1.connect("double", "second_addition")
    pipeline_1.connect("second_addition", "third_addition")

    pipeline_2 = Pipeline(metadata={"type": "another test pipeline", "author": "you"})
    pipeline_2.add_component("first_addition", add_1)
    pipeline_2.add_component("double", Double())
    pipeline_2.add_component("second_addition", add_2)

    pipeline_2.connect("first_addition", "double")
    pipeline_2.connect("double", "second_addition")

    assert marshal_pipelines(pipelines={"pipe1": pipeline_1, "pipe2": pipeline_2}) == {
        "pipelines": {
            "pipe1": {
                "metadata": {"type": "test pipeline", "author": "me"},
                "max_loops_allowed": 100,
                "components": {
                    "first_addition": {
                        "type": "AddFixedValue",
                        "init_parameters": {"add": 200},
                    },
                    "double": {"type": "Double", "init_parameters": {}},
                    "second_addition": {
                        "type": "AddFixedValue",
                        "init_parameters": {},
                    },
                    "third_addition": {"refer_to": "pipe1.first_addition"},
                },
                "connections": [
                    ("first_addition", "double", "value/value"),
                    ("double", "second_addition", "value/value"),
                    ("second_addition", "third_addition", "value/value"),
                ],
            },
            "pipe2": {
                "metadata": {"type": "another test pipeline", "author": "you"},
                "max_loops_allowed": 100,
                "components": {
                    "first_addition": {"refer_to": "pipe1.first_addition"},
                    "double": {"type": "Double", "init_parameters": {}},
                    "second_addition": {"refer_to": "pipe1.second_addition"},
                },
                "connections": [
                    ("first_addition", "double", "value/value"),
                    ("double", "second_addition", "value/value"),
                ],
            },
        },
        "dependencies": ["test", "canals"],
    }


def test_unmarshal():
    pipelines = unmarshal_pipelines(
        {
            "pipelines": {
                "pipe1": {
                    "metadata": {"type": "test pipeline", "author": "me"},
                    "max_loops_allowed": 100,
                    "components": {
                        "first_addition": {
                            "type": "AddFixedValue",
                            "init_parameters": {"add": 300},
                        },
                        "double": {"type": "Double"},
                        "second_addition": {
                            "type": "AddFixedValue",
                            "init_parameters": {},
                        },
                        "third_addition": {"refer_to": "pipe1.first_addition"},
                    },
                    "connections": [
                        ("first_addition", "double", "value/value"),
                        ("double", "second_addition", "value/value"),
                        ("second_addition", "third_addition", "value/value"),
                    ],
                },
                "pipe2": {
                    "metadata": {"type": "another test pipeline", "author": "you"},
                    "max_loops_allowed": 100,
                    "components": {
                        "first_addition": {"refer_to": "pipe1.first_addition"},
                        "double": {"type": "Double"},
                        "second_addition": {"refer_to": "pipe1.second_addition"},
                    },
                    "connections": [
                        ("first_addition", "double", "value/value"),
                        ("double", "second_addition", "value/value"),
                    ],
                },
            },
            "dependencies": ["test", "canals"],
        }
    )

    pipe1 = pipelines["pipe1"]
    assert pipe1.metadata == {"type": "test pipeline", "author": "me"}

    first_addition = pipe1.get_component("first_addition")
    assert type(first_addition) == AddFixedValue
    assert first_addition.defaults["add"] == 300

    second_addition = pipe1.get_component("second_addition")
    assert type(second_addition) == AddFixedValue
    assert second_addition.defaults["add"] == 1
    assert second_addition != first_addition

    third_addition = pipe1.get_component("third_addition")
    assert type(third_addition) == AddFixedValue
    assert third_addition.defaults["add"] == 300
    assert third_addition == first_addition

    double = pipe1.get_component("double")
    assert type(double) == Double

    assert list(pipe1.graph.edges) == [
        ("first_addition", "double", "value/value"),
        ("double", "second_addition", "value/value"),
        ("second_addition", "third_addition", "value/value"),
    ]

    pipe2 = pipelines["pipe2"]
    assert pipe2.metadata == {"type": "another test pipeline", "author": "you"}

    first_addition_2 = pipe2.get_component("first_addition")
    assert type(first_addition_2) == AddFixedValue
    assert first_addition_2.defaults["add"] == 300
    assert first_addition_2 == first_addition

    second_addition_2 = pipe2.get_component("second_addition")
    assert type(second_addition_2) == AddFixedValue
    assert second_addition_2.defaults["add"] == 1
    assert second_addition_2 != first_addition_2
    assert second_addition_2 == second_addition

    with pytest.raises(ValueError):
        pipe2.get_component("third_addition")

    double_2 = pipe2.get_component("double")
    assert type(double_2) == Double
    assert double_2 != double

    assert list(pipe2.graph.edges) == [
        ("first_addition", "double", "value/value"),
        ("double", "second_addition", "value/value"),
    ]



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
import logging
from pathlib import Path
from pprint import pprint

from canals.pipeline import Pipeline
from test.sample_components import AddFixedValue, Parity, Double, Subtract

logging.basicConfig(level=logging.DEBUG)


def test_pipeline(tmp_path):
    add_one = AddFixedValue()

    pipeline = Pipeline()
    pipeline.add_component("add_one", add_one)
    pipeline.add_component("parity", Parity())
    pipeline.add_component("add_ten", AddFixedValue(add=10))
    pipeline.add_component("double", Double())
    pipeline.add_component("add_four", AddFixedValue(add=4))
    pipeline.add_component("add_two", add_one)
    pipeline.add_component("diff", Subtract())

    pipeline.connect("add_one", "parity")
    pipeline.connect("parity.even", "add_four.value")
    pipeline.connect("parity.odd", "double")
    pipeline.connect("add_ten", "diff.first_value")
    pipeline.connect("double", "diff.second_value")
    pipeline.connect("parity.odd", "add_ten.value")
    pipeline.connect("add_four", "add_two")

    pipeline.draw(tmp_path / "fixed_decision_and_merge_pipeline.png")

    results = pipeline.run(
        {
            "add_one": AddFixedValue.Input(value=1),
            "add_two": AddFixedValue.Input(add=2),
        }
    )
    pprint(results)
    assert results == {"add_two": AddFixedValue.Output(value=8)}

    results = pipeline.run(
        {
            "add_one": AddFixedValue.Input(value=2),
            "add_two": AddFixedValue.Input(add=2),
        }
    )
    pprint(results)
    assert results == {"diff": Subtract.Output(difference=7)}


if __name__ == "__main__":
    test_pipeline(Path(__file__).parent)



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
from pathlib import Path
from pprint import pprint

from canals.pipeline import Pipeline
from test.sample_components import AddFixedValue, Parity, Double

import logging

logging.basicConfig(level=logging.DEBUG)


def test_pipeline(tmp_path):
    add_one = AddFixedValue(add=1)

    pipeline = Pipeline()
    pipeline.add_component("add_one", add_one)
    pipeline.add_component("parity", Parity())
    pipeline.add_component("add_ten", AddFixedValue(add=10))
    pipeline.add_component("double", Double())
    pipeline.add_component("add_three", AddFixedValue(add=3))

    pipeline.connect("add_one", "parity")
    pipeline.connect("parity.even", "add_ten.value")
    pipeline.connect("parity.odd", "double.value")
    pipeline.connect("add_ten", "add_three")

    pipeline.draw(tmp_path / "fixed_decision_pipeline.png")

    results = pipeline.run({"add_one": AddFixedValue.Input(value=1)})
    pprint(results)
    assert results == {"add_three": AddFixedValue.Output(value=15)}

    results = pipeline.run({"add_one": AddFixedValue.Input(value=2)})
    pprint(results)
    assert results == {"double": Double.Output(value=6)}


if __name__ == "__main__":
    test_pipeline(Path(__file__).parent)



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
from dataclasses import dataclass

import pytest

from canals.component import component, ComponentInput, ComponentOutput
from canals.testing import BaseTestComponent


@component
class Double:
    """
    Doubles the input value.
    """

    @dataclass
    class Input(ComponentInput):
        value: int

    @dataclass
    class Output(ComponentOutput):
        value: int

    def run(self, data: Input) -> Output:
        """
        Doubles the input value
        """
        return Double.Output(value=data.value * 2)


import pytest
from canals.testing import BaseTestComponent


class TestDouble(BaseTestComponent):
    @pytest.fixture
    def components(self):
        return [Double()]

    def test_double_default(self):
        component = Double()
        results = component.run(Double.Input(value=10))
        assert results == Double.Output(value=20)
        assert component.init_parameters == {}



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
from pathlib import Path
from pprint import pprint

from canals.pipeline import Pipeline
from test.sample_components import AddFixedValue, Repeat, Double

import logging

logging.basicConfig(level=logging.DEBUG)


def test_pipeline(tmp_path):
    add_one = AddFixedValue(add=1)

    pipeline = Pipeline()
    pipeline.add_component("add_one", add_one)
    pipeline.add_component("repeat", Repeat(outputs=["first", "second"]))
    pipeline.add_component("add_ten", AddFixedValue(add=10))
    pipeline.add_component("double", Double())
    pipeline.add_component("add_three", AddFixedValue(add=3))
    pipeline.add_component("add_one_again", add_one)

    pipeline.connect("add_one.value", "repeat.value")
    pipeline.connect("repeat.first", "add_ten.value")
    pipeline.connect("repeat.second", "double")
    pipeline.connect("repeat.second", "add_three.value")
    pipeline.connect("add_three", "add_one_again")

    pipeline.draw(tmp_path / "parallel_branches_pipeline.png")

    results = pipeline.run({"add_one": AddFixedValue.Input(value=1)})
    pprint(results)

    assert results == {
        "add_one_again": AddFixedValue.Output(value=6),
        "add_ten": AddFixedValue.Output(value=12),
        "double": Double.Output(value=4),
    }


if __name__ == "__main__":
    test_pipeline(Path(__file__).parent)



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
from test.sample_components.test_accumulate import Accumulate
from test.sample_components.test_add_value import AddFixedValue
from test.sample_components.test_double import Double
from test.sample_components.test_parity import Parity
from test.sample_components.test_greet import Greet
from test.sample_components.test_merge_loop import MergeLoop
from test.sample_components.test_remainder import Remainder
from test.sample_components.test_repeat import Repeat
from test.sample_components.test_subtract import Subtract
from test.sample_components.test_sum import Sum
from test.sample_components.test_threshold import Threshold



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
class PipelineError(Exception):
    pass


class PipelineRuntimeError(Exception):
    pass


class PipelineConnectError(PipelineError):
    pass


class PipelineValidationError(PipelineError):
    pass


class PipelineMaxLoops(PipelineError):
    pass


class ComponentError(Exception):
    pass



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
from typing import Literal, Optional, Dict, get_args, Any

import logging
from pathlib import Path

import networkx

from canals.pipeline._utils import find_pipeline_inputs, find_pipeline_outputs
from canals.draw.graphviz import to_agraph
from canals.draw.mermaid import to_mermaid_image, to_mermaid_text


logger = logging.getLogger(__name__)
RenderingEngines = Literal["graphviz", "mermaid-img", "mermaid-text"]


def draw(
    graph: networkx.MultiDiGraph,
    path: Path,
    engine: RenderingEngines = "mermaid-img",
    style_map: Optional[Dict[str, str]] = None,
) -> None:
    """
    Renders the pipeline graph and saves it to file.
    """
    converted_graph = convert(graph=graph, engine=engine, style_map=style_map)

    if engine == "graphviz":
        converted_graph.draw(path)

    elif engine == "mermaid-img":
        with open(path, "wb") as imagefile:
            imagefile.write(converted_graph)

    elif engine == "mermaid-text":
        with open((path), "w", encoding="utf-8") as textfile:
            textfile.write(converted_graph)

    else:
        raise ValueError(f"Unknown rendering engine '{engine}'. Choose one from: {get_args(RenderingEngines)}.")

    logger.debug("Pipeline diagram saved at %s", path)


def convert(
    graph: networkx.MultiDiGraph,
    engine: RenderingEngines = "mermaid-img",
    style_map: Optional[Dict[str, str]] = None,
) -> Any:
    """
    Renders the pipeline graph with the correct render and returns it.
    """
    graph = _prepare_for_drawing(graph=graph, style_map=style_map or {})

    if engine == "graphviz":
        return to_agraph(graph=graph)

    if engine == "mermaid-img":
        return to_mermaid_image(graph=graph)

    if engine == "mermaid-text":
        return to_mermaid_text(graph=graph)

    raise ValueError(f"Unknown rendering engine '{engine}'. Choose one from: {get_args(RenderingEngines)}.")


def _prepare_for_drawing(graph: networkx.MultiDiGraph, style_map: Dict[str, str]) -> networkx.MultiDiGraph:
    """
    Prepares the graph to be drawn: adds explitic input and output nodes, labels the edges, applies the styles, etc.
    """
    # Apply the styles
    if style_map:
        for node, style in style_map.items():
            graph.nodes[node]["style"] = style

    # Label the edges
    for inp, outp, key, data in graph.edges(keys=True, data=True):
        data["label"] = f"{data['from_socket'].name} -> {data['to_socket'].name}"
        graph.add_edge(inp, outp, key=key, **data)

    # Draw the inputs
    graph.add_node("input")
    for node, in_sockets in find_pipeline_inputs(graph).items():
        for in_socket in in_sockets:
            node_instance = graph.nodes[node]["instance"]
            input_node_defaults = hasattr(node_instance, "defaults") and in_socket.name in node_instance.defaults
            if not input_node_defaults and not (graph.nodes[node]["variadic_input"] and not graph.in_edges(node)):
                graph.add_edge("input", node, label=in_socket.name)

    # Draw the outputs
    graph.add_node("output")
    for node, out_sockets in find_pipeline_outputs(graph).items():
        for out_socket in out_sockets:
            graph.add_edge(node, "output", label=out_socket.name)

    return graph



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
from canals.component.component import component
from canals.component.decorators import save_init_params
from canals.component.input_output import ComponentInput, ComponentOutput, VariadicComponentInput



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
from typing import Union, Callable, Optional
import sys
import builtins
from importlib import import_module
from dataclasses import dataclass

import pytest
from canals.component import component, ComponentInput, ComponentOutput
from canals.testing import BaseTestComponent


@component
class Accumulate:
    """
    Accumulates the value flowing through the connection into an internal attribute.
    The sum function can be customized.

    Example of how to deal with serialization when some of the parameters
    are not directly serializable.
    """

    @dataclass
    class Input(ComponentInput):
        value: int

    @dataclass
    class Output(ComponentOutput):
        value: int

    def __init__(self, function: Optional[Union[Callable, str]] = None):
        """
        :param function: the function to use to accumulate the values.
            The function must take exactly two values.
            If it's a callable, it's used as it is.
            If it's a string, the component will look for it in sys.modules and
            import it at need. This is also a parameter.
        """
        self.state = 0

        if function is None:
            self.function = lambda x, y: x + y
        else:
            self.function = self._load_function(function)
            # 'function' is not serializable by default, so we serialize it manually.
            self.init_parameters = {"function": self._save_function(function)}

    def run(self, data: Input) -> Output:
        self.state = self.function(self.state, data.value)
        return Accumulate.Output(value=self.state)

    def _load_function(self, function: Union[Callable, str]):
        """
        Loads the function by trying to import it.
        """
        if not isinstance(function, str):
            return function

        parts = function.split(".")
        module_name = ".".join(parts[:-1])
        function_name = parts[-1]

        module = import_module(module_name)
        return getattr(module, function_name)

    def _save_function(self, function: Union[Callable, str]):
        """
        Saves the function by returning its import path to be used with `_load_function`
        (which uses `import_module` internally).
        """
        if isinstance(function, str):
            return function
        module = sys.modules.get(function.__module__)
        if not module:
            raise ValueError("Could not locate the import module.")
        if module == builtins:
            return function.__name__
        return f"{module.__name__}.{function.__name__}"


def my_subtract(first, second):
    return first - second


class TestAccumulate(BaseTestComponent):
    @pytest.fixture
    def components(self):
        return [
            Accumulate(),
            Accumulate(function=my_subtract),
            Accumulate(function="test.sample_components.test_accumulate.my_subtract"),
        ]

    def test_accumulate_default(self):
        component = Accumulate()
        results = component.run(Accumulate.Input(value=10))
        assert results == Accumulate.Output(value=10)
        assert component.state == 10

        results = component.run(Accumulate.Input(value=1))
        assert results == Accumulate.Output(value=11)
        assert component.state == 11

        assert component.init_parameters == {}

    def test_accumulate_callable(self):
        component = Accumulate(function=my_subtract)

        results = component.run(Accumulate.Input(value=10))
        assert results == Accumulate.Output(value=-10)
        assert component.state == -10

        results = component.run(Accumulate.Input(value=1))
        assert results == Accumulate.Output(value=-11)
        assert component.state == -11

        assert component.init_parameters == {
            "function": "test.sample_components.test_accumulate.my_subtract",
        }

    def test_accumulate_string(self):
        component = Accumulate(function="test.sample_components.test_accumulate.my_subtract")

        results = component.run(Accumulate.Input(value=10))
        assert results == Accumulate.Output(value=-10)
        assert component.state == -10

        results = component.run(Accumulate.Input(value=1))
        assert results == Accumulate.Output(value=-11)
        assert component.state == -11

        assert component.init_parameters == {
            "function": "test.sample_components.test_accumulate.my_subtract",
        }



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
from typing import Optional

from dataclasses import dataclass

import pytest

from canals.component import component, ComponentInput, ComponentOutput
from canals.testing.test_component import BaseTestComponent


@component
class AddFixedValue:
    """
    Adds the value of `add` to `value`. If not given, `add` defaults to 1.
    """

    @dataclass
    class Input(ComponentInput):
        value: int
        add: int

    @dataclass
    class Output(ComponentOutput):
        value: int

    def __init__(self, add: Optional[int] = 1):
        if add:
            self.defaults = {"add": add}

    def run(self, data: Input) -> Output:
        return AddFixedValue.Output(value=data.value + data.add)


class TestAddFixedValue(BaseTestComponent):
    @pytest.fixture
    def components(self):
        return [AddFixedValue(), AddFixedValue(add=2)]

    def test_addvalue(self):

        component = AddFixedValue()
        results = component.run(AddFixedValue.Input(value=50, add=10))
        assert results == AddFixedValue.Output(value=60)
        assert component.init_parameters == {}



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
import logging

import networkx
from networkx.drawing.nx_agraph import to_agraph as nx_to_agraph

logger = logging.getLogger(__name__)

try:
    from pygraphviz import AGraph
except ImportError:
    logger.debug(
        "Could not import `pygraphviz`. Please install via: \n"
        "pip install pygraphviz\n"
        "(You might need to run this first: apt install libgraphviz-dev graphviz )"
    )
    AGraph = None


def to_agraph(graph: networkx.MultiDiGraph) -> AGraph:
    """
    Renders a pipeline graph using PyGraphViz. You need to install it and all its system dependencies for it to work.
    """
    if not AGraph:
        raise ImportError(
            "Could not import `pygraphviz`. Please install via: \n"
            "pip install pygraphviz\n"
            "(You might need to run this first: apt install libgraphviz-dev graphviz )"
        )

    for node in graph.nodes:
        if graph.nodes[node].get("style") == "running":
            graph.nodes[node]["color"] = "red"
            graph.nodes[node]["penwidth"] = 3

        if graph.nodes[node].get("style") == "queued":
            graph.nodes[node]["color"] = "#0f0"
            graph.nodes[node]["penwidth"] = 3

    for inp, outp, key, data in graph.out_edges("input", keys=True, data=True):
        data["style"] = "dashed"
        graph.add_edge(inp, outp, key=key, **data)

    for inp, outp, key, data in graph.in_edges("output", keys=True, data=True):
        data["style"] = "dashed"
        graph.add_edge(inp, outp, key=key, **data)

    graph.nodes["input"]["shape"] = "plain"
    graph.nodes["output"]["shape"] = "plain"
    agraph = nx_to_agraph(graph)
    agraph.layout("dot")
    return agraph



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
import logging
from pathlib import Path
from pprint import pprint

from canals.pipeline import Pipeline
from test.sample_components import AddFixedValue, Remainder, Double, Sum

logging.basicConfig(level=logging.DEBUG)


def test_pipeline(tmp_path):
    add_one = AddFixedValue()

    pipeline = Pipeline()
    pipeline.add_component("add_one", add_one)
    pipeline.add_component("parity", Remainder())
    pipeline.add_component("add_ten", AddFixedValue(add=10))
    pipeline.add_component("double", Double())
    pipeline.add_component("add_four", AddFixedValue(add=4))
    pipeline.add_component("add_one_again", add_one)
    pipeline.add_component("sum", Sum())

    pipeline.connect("add_one", "parity")
    pipeline.connect("parity.remainder_is_0", "add_ten.value")
    pipeline.connect("parity.remainder_is_1", "double")
    pipeline.connect("add_one", "sum")
    pipeline.connect("add_ten", "sum")
    pipeline.connect("double", "sum")
    pipeline.connect("parity.remainder_is_1", "add_four.value")
    pipeline.connect("add_four", "add_one_again")
    pipeline.connect("add_one_again", "sum")

    pipeline.draw(tmp_path / "variable_decision_and_merge_pipeline.png")

    results = pipeline.run({"add_one": AddFixedValue.Input(value=1)})
    pprint(results)
    assert results == {"sum": Sum.Output(total=14)}

    results = pipeline.run({"add_one": AddFixedValue.Input(value=2)})
    pprint(results)
    assert results == {"sum": Sum.Output(total=17)}


if __name__ == "__main__":
    test_pipeline(Path(__file__).parent)



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
from typing import Optional

from dataclasses import dataclass
import pytest

from canals.testing import BaseTestComponent
from canals.component import component, ComponentInput, ComponentOutput


@component
class Parity:
    """
    Redirects the value, unchanged, along the 'even' connection if even, or along the 'odd' one if odd.
    """

    @dataclass
    class Input(ComponentInput):
        value: int

    @dataclass
    class Output(ComponentOutput):
        even: Optional[int] = None
        odd: Optional[int] = None

    def run(self, data: Input) -> Output:
        """
        :param value: The value to check for parity
        """
        remainder = data.value % 2
        if remainder:
            return Parity.Output(odd=data.value)
        return Parity.Output(even=data.value)


class TestParity(BaseTestComponent):
    @pytest.fixture
    def components(self):
        return [Parity()]

    def test_parity(self):
        component = Parity()
        results = component.run(Parity.Input(value=1))
        assert results == Parity.Output(odd=1)
        results = component.run(Parity.Input(value=2))
        assert results == Parity.Output(even=2)



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai> SPDX-License-Identifier: Apache-2.0
from typing import Tuple, Optional, List, Iterable, Dict, Any, get_args

import logging
import inspect
import itertools
from dataclasses import dataclass

import networkx

from canals.errors import PipelineConnectError, PipelineValidationError
from canals.component.input_output import fields, ComponentInput


logger = logging.getLogger(__name__)


@dataclass
class OutputSocket:
    name: str
    type: type


@dataclass
class InputSocket:
    name: str
    type: type
    variadic: bool
    taken_by: Optional[str] = None


def parse_connection_name(connection: str) -> Tuple[str, Optional[str]]:
    """
    Returns component-connection pairs from a connect_to/from string
    """
    if "." in connection:
        split_str = connection.split(".", maxsplit=1)
        return (split_str[0], split_str[1])
    return connection, None


def find_input_sockets(component) -> Dict[str, InputSocket]:
    """
    Find a component's input sockets.
    """
    run_signature = inspect.signature(component.__class__.run)

    input_annotation = run_signature.parameters["data"].annotation
    if not input_annotation or input_annotation == inspect.Parameter.empty:
        input_annotation = component.input_type
    variadic = hasattr(input_annotation, "_variadic_component_input")

    input_sockets = {}
    for field in fields(input_annotation):
        # Unwrap List types to get the internal type, if the argument is variadic, and Optionals
        #   Note: we're forced to use type() == type() due to an explicit limitation of the typing library,
        #   that disables `issubclass` on typing classes.
        if variadic or type(field.type) == type(Optional[Any]):  # pylint: disable=unidiomatic-typecheck
            type_ = get_args(field.type)[0]
        else:
            type_ = field.type

        input_sockets[field.name] = InputSocket(name=field.name, type=type_, variadic=variadic)

    return input_sockets


def find_output_sockets(component) -> Dict[str, OutputSocket]:
    """
    Find a component's output sockets.
    """
    run_signature = inspect.signature(component.run)

    return_annotation = run_signature.return_annotation
    if return_annotation == inspect.Parameter.empty:
        return_annotation = component.output_type

    output_sockets = {}
    for field in fields(return_annotation):
        # Unwrap Optionals
        #   Note: we're forced to use type() == type() due to an explicit limitation of the typing library,
        #   that disables `issubclass` on typing classes.
        if type(field.type) == type(Optional[Any]):  # pylint: disable=unidiomatic-typecheck
            type_ = get_args(field.type)[0]
        else:
            type_ = field.type

        output_sockets[field.name] = OutputSocket(name=field.name, type=type_)

    return output_sockets


def connections_status(from_node: str, to_node: str, from_sockets: List[OutputSocket], to_sockets: List[InputSocket]):
    """
    Lists the status of the sockets, for error messages.
    """
    from_sockets_list = "\n".join([f" - {socket.name} ({socket.type.__name__})" for socket in from_sockets])
    to_sockets_list = "\n".join(
        [
            f" - {socket.name} ({socket.type.__name__}, {'taken by '+socket.taken_by if socket.taken_by else 'available'})"
            for socket in to_sockets
        ]
    )
    return f"'{from_node}':\n{from_sockets_list}\n'{to_node}':\n{to_sockets_list}"


def find_unambiguous_connection(
    from_node: str, to_node: str, from_sockets: List[OutputSocket], to_sockets: List[InputSocket]
) -> Tuple[OutputSocket, InputSocket]:
    """
    Find one single possible connection between two lists of sockets.
    """
    # List all combinations of sockets that match by type
    possible_connections = [
        (out_sock, in_sock)
        for out_sock, in_sock in itertools.product(from_sockets, to_sockets)
        if not in_sock.taken_by and out_sock.type == in_sock.type
    ]

    # No connections seem to be possible
    if not possible_connections:
        connections_status_str = connections_status(
            from_node=from_node, from_sockets=from_sockets, to_node=to_node, to_sockets=to_sockets
        )
        raise PipelineConnectError(
            f"Cannot connect '{from_node}' with '{to_node}': "
            f"no matching connections available.\n{connections_status_str}"
        )

    # There's more than one possible connection
    if len(possible_connections) > 1:
        # Try to match by name
        name_matches = [
            (out_sock, in_sock) for out_sock, in_sock in possible_connections if in_sock.name == out_sock.name
        ]
        if len(name_matches) != 1:
            # TODO allow for multiple connections at once if there is no ambiguity?
            # TODO give priority to sockets that have no default values?
            connections_status_str = connections_status(
                from_node=from_node, from_sockets=from_sockets, to_node=to_node, to_sockets=to_sockets
            )
            raise PipelineConnectError(
                f"Cannot connect '{from_node}' with '{to_node}': more than one connection is possible "
                "between these components. Please specify the connection name, like: "
                f"pipeline.connect('component_1.output_value', 'component_2.input_value').\n{connections_status_str}"
            )

    return possible_connections[0]


def find_pipeline_inputs(graph: networkx.MultiDiGraph) -> Dict[str, List[InputSocket]]:
    """
    Collect components that have disconnected input sockets. Note that this method returns *ALL* disconnected
    input sockets, including all such sockets with default values.
    """
    return {
        node: [socket for socket in data.get("input_sockets", {}).values() if not socket.taken_by]
        for node, data in graph.nodes(data=True)
    }


def find_pipeline_outputs(graph) -> Dict[str, List[OutputSocket]]:
    """
    Collect components that have disconnected output sockets. They define the pipeline output.
    """
    return {
        node: list(data.get("output_sockets", {}).values())
        for node, data in graph.nodes(data=True)
        if not graph.out_edges(node)
    }


def _validate_input_sockets_are_connected(graph: networkx.MultiDiGraph, input_values: Dict[str, ComponentInput]):
    """
    Make sure all the inputs nodes are receiving all the values they need, either from the Pipeline's input or from
    other nodes.
    """
    valid_inputs = find_pipeline_inputs(graph)
    for node, sockets in valid_inputs.items():
        if node in input_values:
            for socket in sockets:
                node_instance = graph.nodes[node]["instance"]
                input_in_node_defaults = hasattr(node_instance, "defaults") and socket.name in node_instance.defaults
                if not input_in_node_defaults and not socket.name in input_values[node].names():
                    raise ValueError(f"Missing input: {node}.{socket.name}")


def _validate_nodes_receive_only_expected_input(graph: networkx.MultiDiGraph, input_values: Dict[str, ComponentInput]):
    """
    Make sure that every input node is only receiving input values from EITHER the pipeline's input or another node,
    but never from both.
    """
    for node, input_data in input_values.items():
        for socket_name in input_data.names():
            if not getattr(input_data, socket_name):
                continue
            if not socket_name in graph.nodes[node]["input_sockets"].keys():
                raise ValueError(f"Component {node} is not expecting any input value called {socket_name}")

            taken_by = graph.nodes[node]["input_sockets"][socket_name].taken_by
            if taken_by:
                raise ValueError(f"The input {socket_name} of {node} is already taken by node {taken_by}")


def validate_pipeline_input(  # pylint: disable=too-many-branches
    graph: networkx.MultiDiGraph, input_values: Dict[str, ComponentInput]
) -> Dict[str, ComponentInput]:
    """
    Make sure the pipeline is properly built and that the input received makes sense.
    Returns the input values, validated and updated at need.
    """
    input_components = find_pipeline_inputs(graph)
    if not find_pipeline_inputs(graph):
        raise PipelineValidationError("This pipeline has no inputs.")

    # Make sure the input keys are all nodes of the pipeline
    unknown_components = [key for key in input_values.keys() if not key in graph.nodes]
    if unknown_components:
        raise ValueError(f"Pipeline received data for unknown component(s): {', '.join(unknown_components)}")

    # Make sure all necessary sockets are connected
    _validate_input_sockets_are_connected(graph, input_values)

    # Make sure that the pipeline input is only sent to nodes that won't receive data from other nodes
    _validate_nodes_receive_only_expected_input(graph, input_values)

    # Make sure variadic input components are receiving lists
    for component in input_components.keys():
        if graph.nodes[component]["variadic_input"] and component in input_values.keys():
            for key, value in input_values[component].__dict__.items():  # should be just one
                if not isinstance(value, Iterable):
                    setattr(input_values[component], key, [value])

    return input_values



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
import logging
import base64

import requests
import networkx


logger = logging.getLogger(__name__)


MERMAID_STYLED_TEMPLATE = """
%%{{ init: {{'theme': 'neutral' }} }}%%

{graph_as_text}

style IN  fill:#fff,stroke:#fff,stroke-width:1px
style OUT fill:#fff,stroke:#fff,stroke-width:1px
classDef running fill:#fff,stroke:#f00,stroke-width:2px
classDef queued fill:#fff,stroke:#0f0,stroke-width:2px
linkStyle default stroke-width:2px,stroke-dasharray: 5 5;
{solid_arrows}
"""


def to_mermaid_image(graph: networkx.MultiDiGraph):
    """
    Renders a pipeline using Mermaid (hosted version at 'https://mermaid.ink'). Requires Internet access.
    """
    graph_as_text = to_mermaid_text(graph=graph)
    num_solid_arrows = len(graph.edges) - len(graph.in_edges("output")) - len(graph.out_edges("input"))
    solid_arrows = "\n".join([f"linkStyle {i} stroke-width:2px;" for i in range(num_solid_arrows)])

    graph_styled = MERMAID_STYLED_TEMPLATE.format(graph_as_text=graph_as_text, solid_arrows=solid_arrows)
    logger.debug("Mermaid diagram:\n%s", graph_styled)

    graphbytes = graph_styled.encode("ascii")
    base64_bytes = base64.b64encode(graphbytes)
    base64_string = base64_bytes.decode("ascii")
    url = "https://mermaid.ink/img/" + base64_string

    logging.debug("Rendeding graph at %s", url)
    try:
        resp = requests.get(url, timeout=10)
        if resp.status_code >= 400:
            logger.warning("Failed to draw the pipeline: https://mermaid.ink/img/ returned status %s", resp.status_code)
            logger.info("Exact URL requested: %s", url)
            logger.warning("No pipeline diagram will be saved.")
            return None
    except Exception as exc:  # pylint: disable=broad-except
        logger.warning("Failed to draw the pipeline: could not connect to https://mermaid.ink/img/ (%s)", exc)
        logger.info("Exact URL requested: %s", url)
        logger.warning("No pipeline diagram will be saved.")
        return None

    return resp.content


def to_mermaid_text(graph: networkx.MultiDiGraph) -> str:
    """
    Converts a Networkx graph into Mermaid syntax. The output of this function can be used in the documentation
    with `mermaid` codeblocks and it will be automatically rendered.
    """
    components = {
        comp: f"{comp}:::{graph.nodes[comp]['style']}" if "style" in graph.nodes[comp] else comp for comp in graph.nodes
    }

    connections_list = [
        f"{components[from_comp]} -- {conn_data['label']} --> {components[to_comp]}"
        for from_comp, to_comp, conn_data in graph.edges(data=True)
        if from_comp != "input" and to_comp != "output"
    ]
    input_connections = [
        f"IN([input]) -- {conn_data['label']} --> {components[to_comp]}"
        for _, to_comp, conn_data in graph.out_edges("input", data=True)
    ]
    output_connections = [
        f"{components[from_comp]} -- {conn_data['label']} --> OUT([output])"
        for from_comp, _, conn_data in graph.in_edges("output", data=True)
    ]
    connections = "\n".join(connections_list + input_connections + output_connections)

    mermaid_graph = f"graph TD;\n{connections}"
    return mermaid_graph



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
from pathlib import Path
from pprint import pprint

from canals.pipeline import Pipeline
from test.sample_components import AddFixedValue, Double

import logging

logging.basicConfig(level=logging.DEBUG)


def test_pipeline(tmp_path):
    pipeline = Pipeline()
    pipeline.add_component("first_addition", AddFixedValue(add=2))
    pipeline.add_component("second_addition", AddFixedValue())
    pipeline.add_component("double", Double())
    pipeline.connect("first_addition", "double")
    pipeline.connect("double", "second_addition")

    pipeline.draw(tmp_path / "linear_pipeline.png")

    results = pipeline.run({"first_addition": AddFixedValue.Input(value=1)})
    pprint(results)

    assert results == {"second_addition": AddFixedValue.Output(value=7)}


if __name__ == "__main__":
    test_pipeline(Path(__file__).parent)



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
from typing import *
from pathlib import Path
from pprint import pprint

from canals.pipeline import Pipeline
from test.sample_components import Accumulate, AddFixedValue, Threshold, MergeLoop

import logging

logging.basicConfig(level=logging.DEBUG)


def test_pipeline(tmp_path):
    accumulator = Accumulate()
    merge_loop = MergeLoop(expected_type=int)

    pipeline = Pipeline(max_loops_allowed=10)
    pipeline.add_component("merge", merge_loop)
    pipeline.add_component("below_10", Threshold(threshold=10))
    pipeline.add_component("accumulator", accumulator)
    pipeline.add_component("add_two", AddFixedValue(add=2))

    pipeline.connect("merge", "below_10")
    pipeline.connect("below_10.below", "accumulator")
    pipeline.connect("accumulator", "merge")
    pipeline.connect("below_10.above", "add_two.value")

    pipeline.draw(tmp_path / "looping_pipeline.png")

    results = pipeline.run({"merge": merge_loop.input_type(4)})
    pprint(results)
    print("accumulator: ", accumulator.state)

    assert results == {"add_two": AddFixedValue.Output(value=18)}
    assert accumulator.state == 16


if __name__ == "__main__":
    test_pipeline(Path(__file__).parent)



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
from dataclasses import is_dataclass

import pytest

from canals.pipeline import Pipeline, save_pipelines, load_pipelines, marshal_pipelines

#
# TODO right now the tests iterate on components, so it's not always clear which component fails the tests.
# Can we instead parametrize the tests on the output of the components fixture?
#
class BaseTestComponent:
    """
    Base tests for components.
    """

    @pytest.fixture
    def components(self):
        """
        Provide here a list of differnetly instantiated components to be tested by this suite.
        Should cover most relevant combinations to get a good coverage of your component's basic features.
        """

    def test_component_contract(self, components):
        """
        Very similar checks are done by the decorator itself,
        but for complex components we might mess with these things after instantiation.
        Let's double check it's all good on this front.
        """
        for component in components:
            assert hasattr(component, "__canals_component__")
            assert hasattr(component, "run")
            if hasattr(component, "Output"):
                assert is_dataclass(component.Output)
            else:
                assert hasattr(component, "output_type")

    def test_save_and_load_in_pipeline(self, tmp_path, components):
        """
        Verifies whether the components can be properly serialized and de-serialized.
        """
        for component in components:
            pipe = Pipeline()
            pipe.add_component("component", component)
            save_pipelines({"pipe": pipe}, tmp_path / "test_pipe.json")

            print(marshal_pipelines({"pipe": pipe}))

            loaded_pipes = load_pipelines(tmp_path / "test_pipe.json")
            new_pipe = loaded_pipes["pipe"]

            assert pipe == new_pipe
